package com.bolour.sample.eclipse.service.exponentiation;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExecutableExtension;
import org.eclipse.ui.WorkbenchException;

import com.bolour.sample.eclipse.service.ui.IFunction;

/**
 * Callback class for exponentiation functions. 
 * 
 * <p>
 * This class provides a generic exponentiation function.
 * A particular instance of this class is parameterized by 
 * a specific exponent to be used by that instance in 
 * exponentiation. The exponent is provided to this instance
 * via an XML attribute called <code>constant</code> in its extension 
 * specification.
 * 
 * <p>
 * Because this class implementes the <code>IExecutableExtension</code>
 * interface, the standard callback creation method of Eclipse,
 * namely <code>IConfigurationElement.createExecutableExtension</code>
 * knows to initialize a created instance of the class by calling 
 * the <code>IExecutableExtension</code> method <code>setInitializationData</code>.
 * This method extracts the value of the <i>constant</i> attribute
 * and initializes the instance with its integer value as the 
 * exponent for later computations.
 * 
 * @author Azad
 */
public class Exponentiation implements IFunction, IExecutableExtension {

	private static final String POWER_ATTRIBUTE = "constant";
	
	private int power = 0;
		
	/**
	 * Implementation of the exponentiation operation.
	 */
	public long compute(long x) {		
		long result = 1;
		for (int i = 0; i < power; i++) {
			result *= x;
		}
		return result;
	}

	/**
	 * Initialization method for instances of Exponentiation.
	 * The exponent for this particular listener is provided in the <code>constant</code>
	 * XML attribute of its extension element.
	 * Extract the exponent and remember it for use in later computations.
	 * 
	 * @see org.eclipse.core.runtime.IExecutableExtension#setInitializationData(org.eclipse.core.runtime.IConfigurationElement, java.lang.String, java.lang.Object)
	 */
	public void setInitializationData(IConfigurationElement member, String propertyName, 
	  Object data) throws CoreException {
		
		String s = member.getAttribute(POWER_ATTRIBUTE);
		try {
			power = Integer.parseInt(s);
		}
		catch (NumberFormatException ex) {
			String message = "Exponentiation: unable to convert " 
			  + POWER_ATTRIBUTE + " attribute '" + s + "' to an integer.";
			System.err.println(message);
			throw new WorkbenchException(message);
		}
		if (power < 0) {
			String message = "Exponentiation: negative powers unsupported: " 
			  + power;
			System.err.println(message);
			throw new WorkbenchException(message);
		}
	}
}
