package com.bolour.sample.eclipse.service.ui;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtension;
import org.eclipse.core.runtime.IExtensionPoint;
import org.eclipse.core.runtime.IPluginRegistry;
import org.eclipse.core.runtime.Platform;
import org.eclipse.ui.WorkbenchException;

/**
 * Extension processing logic for the <code>functions</code> extension-point.
 * Extract specific information about each function. Create callback
 * function object when required.
 * 
 * @author Azad
 */
public class ProcessServiceMembers {

	/**
	 * The fully-qualified name of the functions extension-point for this plug-in.
	 */
	private static final String EXTENSION_POINT =
		"com.bolour.sample.eclipse.service.ui.functions";

	/**
	 * Name of the XML attribute designating a function's name.
	 */
	private static final String FUNCTION_NAME_ATTRIBUTE = "name";

	/**
	 * Name of the XML attribute designating the fully-qualified name 
	 * of the implementation class of a function.
	 */
	private static final String CLASS_ATTRIBUTE = "class";

	/**
	 * Name of the XML attribute designating the constant of the 
	 * computation for a function (if any).
	 */
	private static final String CONSTANT_ATTRIBUTE = "constant";

	/**
	 * Perform initial extension processing for the members of the
	 * <code>functions</code> extension-point. Make calls to the user interface
	 * module to add the functions of an extension to the UI functions grid.
	 * For each function, a virtual proxy callback object is created and handed
	 * to the user interface module. The proxy class is a nested top-level 
	 * class and is therefore known at compile time. The actual (real) callback
	 * objects configured into extensions are instantiated and initialized in 
	 * a lazy fashion by the proxy callback objects.
	 * 
	 * @param grid The UI functions grid exposing the functions configured
	 * 		into <code>functions</code> extensions.
	 * 
	 */
	public static void process(FunctionsGrid grid) throws WorkbenchException {
		IPluginRegistry registry = Platform.getPluginRegistry();
		IExtensionPoint extensionPoint = registry.getExtensionPoint(EXTENSION_POINT);
		if (extensionPoint == null) {
			throw new WorkbenchException(
				"unable to resolve extension-point: " + EXTENSION_POINT);
		}
		IConfigurationElement[] members = extensionPoint.getConfigurationElements();

		// For each service:
		for (int m = 0; m < members.length; m++) {
			IConfigurationElement member = members[m];
			// Get the label of the extender plugin and the ID of the extension.
			IExtension extension = member.getDeclaringExtension();
			String pluginLabel = extension.getDeclaringPluginDescriptor().getLabel();
			if (pluginLabel == null) {
				pluginLabel = "[unnamed plugin]";
			}
			// Get the name of the operation implemented by the service.
			// The operation name is a service attribute in the extension's XML specification.
			String functionName = member.getAttribute(FUNCTION_NAME_ATTRIBUTE);
			if (functionName == null) {
				functionName = "[unnamed function]";
			}
			String label = pluginLabel + "/" + functionName;
			Integer constant = null;
			String s = member.getAttribute(CONSTANT_ATTRIBUTE);
			if (s != null) {
				try {
					constant = new Integer(s);
				}
				catch (NumberFormatException ex) {
					// Invalid function. Inform user but ignore.
					System.err.println(
						"unable to convert constant parameter'"
							+ s
							+ "' to integer for function '"
							+ label
							+ "' [ignoring function]");
					continue;
				}
			}
			// Add a row to support this operation to the UI grid.
			IFunction proxy = new FunctionProxy(member);
			grid.addFunction(proxy, functionName, label, constant);
		}
	}

	/**
	 * Virtual proxy class for arithmetic function invokation.
	 * Members of this class support lazy processing by fronting the real 
	 * implementations of arithmetic functions configured into extensions.
	 */
	private static class FunctionProxy implements IFunction {
		// The "real" implementation of the function.
		private IFunction delegate = null;
		// The configured state of the extension element representing this arithmetic function.
		private IConfigurationElement element;
		// Whether this function has been invoked before. 
		// Used for lazy evaluation of the service.
		private boolean invoked = false;

		/**
		 * Construct a virtual proxy to stand in place of an extension function.
		 * The constructor simply keeps track of an arithmetic function's configured state
		 * without at this time instantiating the implementation of the function.
		 * 
		 * @param element The configuration of this arithmetic function.
		 */
		public FunctionProxy(IConfigurationElement element) {
			this.element = element;
		}

		/**
		 * Compute the function value. 
		 * The proxy computation first instantiates the implementation
		 * of the function, if this is the first time the function is called,
		 * and then delegates to that implementation. The instantiation and
		 * initialization of the delegate implementation uses the standard 
		 * callback object instantiation and initialization methods of Eclipse.
		 * 
		 * @see com.bolour.sample.eclipse.service.ui.IFunction#compute(long)
		 */
		public final long compute(long x) throws ArithmeticException {
			try {
				getDelegate();
			}
			catch (Exception ex) {
				throw new ArithmeticException("invalid function");
			}
			if (delegate == null) {
				throw new ArithmeticException("invalid function");
			}
			return delegate.compute(x);
		}

		/**
		 * Instantiate and initialize the implementation of the function if
		 * this is the first time the proxy has been called.
		 * 
		 * @return The implementation delegate (same as the instance variable <code>delegate</code>.
		 * @throws Exception If the callback object is misconfigured.
		 */
		private final IFunction getDelegate() throws Exception {
			if (invoked) {
				return delegate;
			}
			invoked = true;
			try {
				Object callback = element.createExecutableExtension(CLASS_ATTRIBUTE);
				if (!(callback instanceof IFunction)) {
					String message =
						"callback class '"
							+ callback.getClass().getName()
							+ "' is not an IFunction";
					System.err.println(message);
					throw new ClassCastException(message);
				}
				delegate = (IFunction)callback;
			}
			catch (CoreException ex) {
				System.err.println(ex.getMessage());
				ex.printStackTrace();
				throw ex;
			}
			catch (Error err) {
				System.err.println(err.getMessage());
				err.printStackTrace();
				throw err;
			}
			return delegate;
		}
	}
}